/*
 *	The Reinhardt style (version 0.7.1)
 *	Thank you to Chriss Lee <lee@azsites.com>
 *	             Carsten Pfei <pfeiffer@kde.org>
 *	             Karol Szwed <gallium@kde.org>
 *	For the KDE3 dotNET Style (version 1.2) which was used as the basis for
 *	this style. The following disclaimer is the original from dotNET Style (version 1.2)
 *
 *	Drawing routines completely reimplemented from KDE3 HighColor, which was
 *	originally based on some stuff from the KDE2 HighColor.
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License version 2 as published by the Free Software Foundation.
 *
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public License
 *	along with this library; see the file COPYING.LIB.  If not, write to
 *	the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *	Boston, MA 02111-1307, USA.
 */
#include <qstyleplugin.h>
#include <qstylefactory.h>
#include <qpointarray.h>
#include <qpainter.h>
#include <qtabbar.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qcombobox.h>
#include <qlistbox.h>
#include <qlistview.h>
#include <qscrollbar.h>
#include <qpushbutton.h>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qprogressbar.h>
#include <qdrawutil.h>
#include <qapplication.h>
#include <qvariant.h>
#include <qpixmapcache.h>
#include <qslider.h>
#include <qsettings.h>

#include <kpixmap.h>

#include "reinhardt.h"

// -- Style Plugin Interface -------------------------
class reinhardtPlugin : public QStylePlugin
{
	public:
		reinhardtPlugin() {}
		~reinhardtPlugin() {}

		QStringList keys() const {
			return QStringList() << "reinhardt";
		}

		QStyle* create( const QString& key ) {
			if ( key == "reinhardt" )
				return new reinhardt;
			return 0;
		}
};

Q_EXPORT_PLUGIN( reinhardtPlugin )

// ---------------------------------------------------

reinhardt::reinhardt() : KStyle( AllowMenuTransparency ), kickerMode(false)
{
	winstyle = QStyleFactory::create("Windows");
	if(!winstyle) {
	//	We don't have the Windows style, neither builtin nor as a plugin.
	//	Use any style rather than crashing.
		winstyle = QStyleFactory::create(*(QStyleFactory::keys().begin()));
	}

	if (qApp->inherits("KApplication")) {
		connect( qApp, SIGNAL( kdisplayPaletteChanged() ), SLOT( paletteChanged() ));
	}

	settings = new QSettings();
	useTextShadows = settings->readBoolEntry("/reinhardtstyle/Settings/useTextShadows", true);
	useDropShadows = settings->readBoolEntry("/reinhardtstyle/Settings/useDropShadows", true);
	useHighlightedButtons = settings->readBoolEntry("/reinhardtstyle/Settings/useHighlightedButtons", true);
	useOldTabs = settings->readBoolEntry("/reinhardtstyle/Settings/useOldTabs", false);
	useOldCombo = settings->readBoolEntry("/reinhardtstyle/Settings/useOldCombo", false);
	useOldTreeview = settings->readBoolEntry("/reinhardtstyle/Settings/useOldTreeview", false);
	reverseLayout = QApplication::reverseLayout();
}


reinhardt::~reinhardt()
{
}

bool reinhardt::inheritsKHTML(const QWidget* w) const
{
	if (w->parentWidget(true) && w->parentWidget(true)->parentWidget(true) && w->parentWidget(true)->parentWidget(true)->parentWidget(true) &&  w->parentWidget(true)->parentWidget(true)->parentWidget(true)->inherits("KHTMLView")) {
		return true;
	} else {
		return false;
	}
}

void reinhardt::polish(QWidget* widget)
{
	if (!qstrcmp(qApp->argv()[0], "kicker") || widget->inherits("Kicker"))
		kickerMode = true;

	if (widget->isTopLevel())
		return;

// we can't simply set a palette -- upon color-theme changes, we have
// to update the palette again.
	bool extraPalette = false;

	if (widget->inherits("QComboBox") && !inheritsKHTML(widget)) {
		widget->installEventFilter (this);
		updatePalette( (QComboBox*) widget );
		extraPalette = true;
	} else if (widget->inherits("QButton") && !inheritsKHTML(widget)) {
		widget->installEventFilter (this);
		updatePalette( (QButton*) widget );
		extraPalette = true;
	} else if (widget->inherits("QCheckBox") && !inheritsKHTML(widget)) {
		widget->installEventFilter (this);
		updatePalette( (QCheckBox*) widget );
		extraPalette = true;
	} else if (widget->inherits("QRadioButton") && !inheritsKHTML(widget)) {
		widget->installEventFilter (this);
		updatePalette( (QRadioButton*) widget );
		extraPalette = true;
	} else {
		winstyle->polish(widget);
	}

//	This code is disabled until Carsten explains to me why it's supposed to
//	be here. It breaks dynamically changing the color from KControl and does
//	other bad things (see bug #54569)
/*
	if (widget->ownPalette()) {
		if ( widget->inherits("QToolBar")) {
			updatePalette( (QToolBar*) widget );
			extraPalette = true;
		} else if (widget->inherits("QMenuBar")) {
			updatePalette( (QMenuBar*) widget );
			extraPalette = true;
		}
	}
*/
}

void reinhardt::unPolish(QWidget* widget)
{
	winstyle->unPolish(widget);

	if (widget->inherits("QComboBox") && !inheritsKHTML(widget)) {
		widget->removeEventFilter (this);
	} else if (widget->inherits("QButton") && !inheritsKHTML(widget)) {
		widget->removeEventFilter (this);
	} else if (widget->inherits("QCheckBox") && !inheritsKHTML(widget)) {
		widget->removeEventFilter (this);
	} else if (widget->inherits("QRadioButton") && !inheritsKHTML(widget)) {
		widget->removeEventFilter (this);
	}
}

void reinhardt::renderMenuBlendPixmap( KPixmap &pix, const QColorGroup &cg, const QPopupMenu *popup ) const
{
	QPainter p( &pix );
	if (QApplication::reverseLayout()) {
		p.fillRect( popup->frameRect().width()-22, 0, 22, pix.height(), cg.mid() );
		p.fillRect( 0, 0, popup->frameRect().width()-22, pix.height(), cg.background().light() );
	} else {
		p.fillRect( 0, 0, pix.width(), pix.height(), cg.background().light() );
		p.fillRect( popup->frameRect().left() + 1, 0, 22, pix.height(), cg.mid() );
	}
}

void reinhardt::renderButton(QPainter *p,
                               const QRect &r,
                               const QColorGroup &g) const
{
	p->save();
	//int x, y, w, h;
//	Quick way to set x,y and w,h
	//r.rect(&x, &y, &w, &h);
	//r.setRect(x + 1, y + 1, w - 2, h - 2);

	p->setPen(g.background().dark());
	//p->setBrush(g.background());
	p->setBrush(g.button());
	p->drawRect(r);

	p->setPen(g.background());
	p->drawPoint(r.topLeft());
	p->drawPoint(r.topRight());
	p->drawPoint(r.bottomLeft());
	p->drawPoint(r.bottomRight());

	p->restore();
}

void reinhardt::renderSlider(QPainter *p,
                               const QRect &r,
                               const QColorGroup &g) const
{
	int x, y, w, h;

	r.rect(&x, &y, &w, &h);

	p->save();
	p->fillRect(x, y, w, h, g.background().dark(115));
	QRect cr(x + 2, y + 2, w - 4 , h - 4);
	p->fillRect(cr, g.button());
	p->setPen(g.button().dark(130));
	p->drawRect(cr);

	p->setPen(g.background().dark(115));
	p->drawPoint(cr.topLeft());
	p->drawPoint(cr.topRight());
	p->drawPoint(cr.bottomLeft());
	p->drawPoint(cr.bottomRight());

	p->restore();
}

void reinhardt::renderPanel(QPainter *p,
                              const QRect &r,
                              const QColorGroup &g,
                              bool sunken) const
{
	int x, x2, y, y2, w, h;
	r.rect(&x,&y,&w,&h);
	r.coords(&x, &y, &x2, &y2);

	p->save();

	if (kickerMode) {
		p->setPen(g.background().dark());
		p->setBrush(g.button());
		p->drawRect(r);

		p->setPen(g.background());
		p->drawPoint(r.topLeft());
		p->drawPoint(r.topRight());
		p->drawPoint(r.bottomLeft());
		p->drawPoint(r.bottomRight());
		p->restore();
		return;
	}
	if (sunken) {
		p->setPen(g.background().light());
		p->drawRect(x, y, w, h);
		p->setPen(g.mid());
		p->drawLine(x, y, x + w, y);
		p->drawLine(x, y, x, y + h);
	}  else {
		p->setPen(g.background().dark());
		p->drawRect(x, y, w, h);
	}
	p->restore();
}


void reinhardt::drawKStylePrimitive(KStylePrimitive kpe,
                                      QPainter *p,
                                      const QWidget* widget,
                                      const QRect &r,
                                      const QColorGroup &cg,
                                      SFlags flags,
                                      const QStyleOption& opt) const
{
//  SLIDER
//  ------
	switch( kpe ) {
		case KPE_SliderGroove: {
			const QSlider* slider = (const QSlider*)widget;
			int x, y, z = 8, w, h;
			r.rect(&x, &y, &w, &h);
			bool horizontal = slider->orientation() == Horizontal;
			int gcenter = (horizontal ? h : w) - 7;

			int swidth = slider->maxValue() - slider->minValue();
			int svalue = slider->value() - slider->minValue();
			if (swidth == 0) swidth = 1; //Can't divide by zero!

			QPen pen(cg.dark(), 5, SolidLine, RoundCap, RoundJoin);
			QPen pen2(cg.mid(), 5, SolidLine, RoundCap, RoundJoin);
			if (horizontal) {
				p->setPen (pen);
				do {
					if (z - 8 >= ((w - 10) * svalue / swidth)) p->setPen(pen2);
					p->drawLine(x + z, gcenter, x + z, 3);
					z = z + 6;
				} while (z < (w - 5));

			} else {
				p->setPen (pen2);
				do {
					if (z - 8 >= ((h - 10) * svalue / swidth)) p->setPen(pen);
					p->drawLine(gcenter, y + z, 3, y + z);
					z = z + 6;
				} while (z < (h - 5));
			}
			break;
		}

		case KPE_SliderHandle: {
			int x, y, w, h;
			r.rect(&x, &y, &w, &h);
			if (h < w) { //vertical slide

				int x = (w - 8), y = r.center().y();
				const QCOORD anarrow[] = { x, y, x + 6, y - 6, x + 6, y + 6, x, y};
				const QPointArray a(4, anarrow);

				p->setPen(flags & Style_Down ? cg.highlight() : cg.text());
				p->setBrush(flags & Style_Down ? cg.highlight() : cg.text());
				p->drawPolygon(a);
			} else {
				int x = r.center().x(), y = (h - 8);
				const QCOORD anarrow[] = { x, y, x + 6, y + 6, x - 6, y + 6, x, y};
				const QPointArray a(4, anarrow);

				p->setPen(flags & Style_Down ? cg.highlight() : cg.text());
				p->setBrush(flags & Style_Down ? cg.highlight() : cg.text());
				p->drawPolygon(a);
			}
			break;
		}

		default:
			KStyle::drawKStylePrimitive(kpe, p, widget, r, cg, flags, opt);
	}
}

// This function draws primitive elements as well as their masks.
void reinhardt::drawPrimitive(PrimitiveElement pe,
                                QPainter *p,
                                const QRect &r,
                                const QColorGroup &cg,
                                SFlags flags,
                                const QStyleOption &opt ) const
{
	bool down   = flags & Style_Down;
	bool on     = flags & Style_On;
	bool sunken = flags & Style_Sunken;
	bool horiz  = flags & Style_Horizontal;

	int x, x2, y, y2, w, h;
	r.rect(&x, &y, &w, &h);
	r.coords(&x, &y, &x2, &y2);

	switch(pe) {
	//	BUTTONS
	//	-------
		case PE_ButtonBevel:
		case PE_ButtonTool:
		case PE_ButtonDropDown:
		case PE_HeaderSection:
		case PE_ButtonCommand: {

			p->save();
			if (on || down) {
				renderButton(p, r, cg);
				p->fillRect(x + 1, y + 1, w - 2, h - 2, cg.highlight());
				p->setPen(cg.highlight().dark(150));
				p->drawRect(x + 1, y + 1, w - 2, h - 2);
				if (useDropShadows) {
					p->setPen(cg.highlight().dark(110));
					p->drawRect(x + 2, y + 2, w - 4, h - 4);
				}
			} else {
				renderButton(p, r, cg);
			}

			p->restore();
			break;
		}

		case PE_ButtonDefault: {
			p->save();
			p->setPen(cg.background().dark(115));
			p->drawLine(x + 3, y, x + w - 5, y);
			p->drawLine(x, y + 3, x, y + h - 5);
			p->drawLine(x + 3, h - 1, w - 5, h - 1);
			p->drawLine(w - 1, y + 3, w - 1, h - 5);
			p->restore();
			break;
		}
		
		case PE_ScrollBarSlider: {
			renderSlider(p, r, cg);
			break;
		}

		case PE_ScrollBarAddPage:
		case PE_ScrollBarSubPage: {
			if (on || down) {
				p->fillRect(r, QBrush(cg.background().dark(180)));
			} else {
				p->fillRect(r, QBrush(cg.background().dark(115)));
			}
			break;
		}

	//	SCROLLBAR BUTTONS
	//	-----------------
		case PE_ScrollBarSubLine: {
			p->fillRect(r, cg.button().light());
			p->setPen(cg.button().dark());
			if (horiz) {
				drawPrimitive(PE_ArrowLeft, p, r, cg, flags);
				p->drawRect(x, y, w, h);
				p->setPen(cg.background());
				p->drawPoint(x, y);
				p->drawPoint(x, y + h - 1);
				if (useDropShadows) {
					p->setPen(cg.button().light(70));
					p->drawLine(x + w - 2, y + 1, x + w - 2, y + h - 2);
					p->setPen(cg.button().light(86));
					p->drawLine(x + w - 3, y + 2, x + w - 3, y + h - 3);
					p->setPen(cg.button().light(97));
					p->drawLine(x + w - 4, y + 2, x + w - 4, y + h - 3);
				}
			} else {
				drawPrimitive(PE_ArrowUp, p, r, cg, flags);
				p->drawRect(x, y, w, h);
				p->setPen(cg.background());
				p->drawPoint(x, y);
				p->drawPoint(x + w - 1, y);
				if (useDropShadows) {
					p->setPen(cg.button().light(70));
					p->drawLine(x + 1, y + h - 2, x + w - 2, y + h - 2);
					p->setPen(cg.button().light(86));
					p->drawLine(x + 2, y + h - 3, x + w - 3, y + h - 3);
					p->setPen(cg.button().light(97));
					p->drawLine(x + 2, y + h - 4, x + w - 3, y + h - 4);
				}
			}
			break;
		}

		case PE_ScrollBarAddLine: {
			p->fillRect(r, cg.button().light());
			p->setPen(cg.button().dark());
			if (horiz) {
				drawPrimitive(PE_ArrowRight, p, r, cg, flags);
				p->drawRect(x, y, w, h);
				p->setPen(cg.background());
				p->drawPoint(x + w - 1, y);
				p->drawPoint(x + w - 1, y + h - 1);
				if (useDropShadows) {
					p->setPen(cg.button().light(70));
					p->drawLine(x + 1, y + 1, x + 1, y + h - 2);
					p->setPen(cg.button().light(86));
					p->drawLine(x + 2, y + 2, x + 2, y + h - 3);
					p->setPen(cg.button().light(97));
					p->drawLine(x + 3, y + 2, x + 3, y + h - 3);
				}
			} else {
				drawPrimitive(PE_ArrowDown, p, r, cg, flags);
				p->drawRect(x, y, w, h);
				p->setPen(cg.background());
				p->drawPoint(x, y + h - 1);
				p->drawPoint(x + w - 1, y + h - 1);
				if (useDropShadows) {
					p->setPen(cg.button().light(70));
					p->drawLine(x + 1, y + 1, x + w - 2, y + 1);
					p->setPen(cg.button().light(86));
					p->drawLine(x + 2, y + 2, x + w - 3, y + 2);
					p->setPen(cg.button().light(97));
					p->drawLine(x + 2, y + 3, x + w - 3, y + 3);
				}
			}
			break;
		}

	//	CHECKBOXES
	//	----------
		case PE_Indicator: {
			p->setPen(cg.background().dark());
			p->setBrush(flags & Style_Enabled ? cg.light() : cg.background());
			p->drawRect(r);
			const QCOORD corners[] = { x, y, x, y2, x2, y, x2, y2 };
			p->setPen(cg.background());
			p->drawPoints(QPointArray(4, corners));
			break;
		}

		case PE_IndicatorMask: {
			p->fillRect (r, color1);
			p->setPen (color0);
			p->drawPoint(r.topLeft());
			p->drawPoint(r.topRight());
			p->drawPoint(r.bottomLeft());
			p->drawPoint(r.bottomRight());
			break;
		}

	//	RADIOBUTTONS
	//	------------
		case PE_ExclusiveIndicator: {
			const QCOORD outline[] = { 0, 7, 0, 5, 1, 4, 1, 3, 2, 2, 3, 1, 4, 1, 5, 0, 7, 0, 8, 1, 9, 1, 10, 2, 11, 3, 11, 4, 12, 5, 12, 7, 11, 8, 11, 9, 10, 10, 9, 11, 8, 11, 7, 12, 5, 12, 4, 11, 3, 11, 2, 10, 1, 9, 1, 8 };

			p->fillRect(r, cg.background());
			p->setPen(cg.background().dark());
			p->setBrush(flags & Style_Enabled ? cg.light() : cg.background());
			p->drawPolygon(QPointArray(28, outline));
			break;
		}

		case PE_ExclusiveIndicatorMask: {
			const QCOORD outline[] = { 0, 7, 0, 5, 1, 4, 1, 3, 2, 2, 3, 1, 4, 1, 5, 0, 7, 0, 8, 1, 9, 1, 10, 2, 11, 3, 11, 4, 12, 5, 12, 7, 11, 8, 11, 9, 10, 10, 9, 11, 8, 11, 7, 12, 5, 12, 4, 11, 3, 11, 2, 10, 1, 9, 1, 8 };
			p->fillRect(r, color0);
			p->setPen(color1);
			p->setBrush(color1);
			p->translate(r.x(), r.y());
			p->drawPolygon(QPointArray(28, outline));
			p->translate(-r.x(), -r.y());
			break;
		}

	//	GENERAL PANELS
	//	--------------
		case PE_Splitter: {
			p->fillRect(r, cg.background());
			QPen pen(cg.mid(), 5, SolidLine, RoundCap, RoundJoin);

			/* 
			OK, some stupid stuff going on with layouts is causing us to have to draw this in both directions... Oh well
			at least amaroK now looks right ;)
			*/
			p->setPen(cg.mid());
			//horizontal lines
			p->drawLine(r.center().x(), 2, r.center().x(), r.center().y() - 11);
			p->drawLine(r.center().x(), r.center().y() + 11, r.center().x(), r.height() - 3);
			//vertical lines
			p->drawLine(2, r.center().y(), r.center().x() - 11, r.center().y());
			p->drawLine(r.center().x() + 11, r.center().y(), r.width() - 3, r.center().y());
			
			p->setPen(pen);
			//center dot
			p->drawLine(r.center().x(), r.center().y(), r.center().x(), r.center().y());
			//horizontal dots
			p->drawLine(r.center().x(), r.center().y() - 6, r.center().x(), r.center().y() - 6);
			p->drawLine(r.center().x(), r.center().y() + 6, r.center().x(), r.center().y() + 6);
			//vertical dots
			p->drawLine(r.center().x() - 6, r.center().y(), r.center().x() - 6, r.center().y());
			p->drawLine(r.center().x() + 6, r.center().y(), r.center().x() + 6, r.center().y());
			break;
		}

		case PE_PanelGroupBox:
		case PE_FocusRect: {
			//don't draw frame is the width is 0
			if ( opt.isDefault() || opt.lineWidth() <= 0 )
				break;
			
			p->setPen(cg.mid());
			p->setBrush(NoBrush);
			p->drawRect(r);

			const QCOORD corners[] = { x, y, x, y2, x2, y, x2, y2 };
			p->setPen(cg.mid().light());
			p->drawPoints(QPointArray(4, corners));

			break;
		}

		case PE_GroupBoxFrame: {
			//don't draw frame is the width is 0
			if ( opt.isDefault() || opt.lineWidth() <= 0 )
				break;
			
			p->setPen(cg.background().dark());
			p->setBrush(NoBrush);
			p->drawRect(r);

			const QCOORD corners[] = { x, y, x, y2, x2, y, x2, y2 };
			p->setPen(cg.background());
			p->drawPoints(QPointArray(4, corners));

			break;
		}

		case PE_StatusBarSection: {
			p->setPen(cg.mid());
			p->drawLine(x - 1, y, x2, y);
			p->drawLine(x - 2, y + 1, x - 2, y2 - 1);
			break;
		}
		
		case PE_WindowFrame: { break; }
		
		case PE_PanelLineEdit:
		case PE_Panel: {
			renderPanel(p, r, cg, sunken);
			break;
		}

		case PE_TabBarBase:	// Still not sure what this one does
		case PE_PanelTabWidget: {
			QRect r2(x, y + 1, w, h - 2);
			renderPanel(p, r2, cg, sunken);
			break;
		}

		case PE_PanelPopup: {
			p->setPen(cg.shadow());
			p->drawRect(r);
			break;
		}

	//	MENU / TOOLBAR PANEL
	//	--------------------
		case PE_PanelMenuBar:			// Menu
		case PE_PanelDockWindow: {		// Toolbar
			p->save();
			p->setPen(cg.button());
			p->setBrush(cg.button());
			p->drawRect(r);
			p->restore();
			break;
		}

	//	TOOLBAR/DOCK WINDOW HANDLE
	//	--------------------------

		case PE_DockWindowHandle:
		case PE_DockWindowResizeHandle: {
			p->fillRect(r, cg.button());
			QPen pen(cg.button().dark(150), 5, SolidLine, RoundCap, RoundJoin);

			if (horiz) {
				if (r.height() < 11) { //If there is only room for one dot, draw only one
					p->setPen(cg.button().dark(150));
					p->drawLine(r.center().x(), 2, r.center().x(), r.center().y() - 3);
					p->drawLine(r.center().x(), r.center().y() + 3, r.center().x(), r.height() - 3);
					p->setPen(pen);
					p->drawLine(r.center().x(), r.center().y(), r.center().x(), r.center().y());
				} else if (r.height() < 18) { //if there is only room for two dots, draw only two
					p->setPen(cg.button().dark(150));
					p->drawLine(r.center().x(), 2, r.center().x(), r.center().y() - 6);
					p->drawLine(r.center().x(), r.center().y() + 6, r.center().x(), r.height() - 3);
					p->setPen(pen);
					p->drawLine(r.center().x(), r.center().y() - 3, r.center().x(), r.center().y() - 3);
					p->drawLine(r.center().x(), r.center().y() + 3, r.center().x(), r.center().y() + 3);
				} else {
					p->setPen(cg.button().dark(150));
					p->drawLine(r.center().x(), 2, r.center().x(), r.center().y() - 11);
					p->drawLine(r.center().x(), r.center().y() + 11, r.center().x(), r.height() - 3);
					p->setPen(pen);
					p->drawLine(r.center().x(), r.center().y() - 6, r.center().x(), r.center().y() - 6);
					p->drawLine(r.center().x(), r.center().y(), r.center().x(), r.center().y());
					p->drawLine(r.center().x(), r.center().y() + 6, r.center().x(), r.center().y() + 6);
				}
			} else {
				if (r.width() < 11) { //if there is only room for one dot, draw only one
					p->setPen(cg.button().dark(150));
					p->drawLine(2, r.center().y(), r.center().x() - 3, r.center().y());
					p->drawLine(r.center().x() + 3, r.center().y(), r.width() - 3, r.center().y());
					p->setPen(pen);
					p->drawLine(r.center().x(), r.center().y(), r.center().x(), r.center().y());
				} else if (r.width() < 18) {
					p->setPen(cg.button().dark(150));
					p->drawLine(2, r.center().y(), r.center().x() - 6, r.center().y());
					p->drawLine(r.center().x() + 6, r.center().y(), r.width() - 3, r.center().y());
					p->setPen(pen);
					p->drawLine(r.center().x() - 3, r.center().y(), r.center().x() - 3, r.center().y());
					p->drawLine(r.center().x() + 3, r.center().y(), r.center().x() + 3, r.center().y());
				} else {
					p->setPen(cg.button().dark(150));
					p->drawLine(2, r.center().y(), r.center().x() - 11, r.center().y());
					p->drawLine(r.center().x() + 11, r.center().y(), r.width() - 3, r.center().y());
					p->setPen(pen);
					p->drawLine(r.center().x() - 6, r.center().y(), r.center().x() - 6, r.center().y());
					p->drawLine(r.center().x(), r.center().y(), r.center().x(), r.center().y());
					p->drawLine(r.center().x() + 6, r.center().y(), r.center().x() + 6, r.center().y());
				}
			}

			break;
		}

	//	TOOLBAR SEPARATOR
	//	-----------------
		case PE_DockWindowSeparator: {
			p->fillRect(r, cg.button());
			p->setPen(cg.button().dark(130));
			if (horiz) {
				p->drawLine(r.center().x(), 2, r.center().x(), r.height() - 3);
			} else {
				p->drawLine(2, r.center().y(), r.width() - 3, r.center().y());
			}
			break;
		}

		case PE_CheckMark: {
			int x = r.center().x() - 3, y = r.center().y() - 3;
			const QCOORD check[] = { x, y + 2, x, y + 4, x + 2, y + 6, x + 6, y + 2, x + 6, y, x + 2, y + 4 };
			const QPointArray a(6, check);

			p->setPen(flags & Style_Down ? cg.highlight() : cg.text());
			p->setBrush(flags & Style_Down ? cg.highlight() : cg.text());
			p->drawPolygon(a);

			break;
		}

		case PE_HeaderArrow:
		case PE_ArrowUp:
		case PE_ArrowDown:
		case PE_ArrowLeft:
		case PE_ArrowRight: {
			QPointArray a;

			switch (pe) {
				case PE_ArrowUp: {
					a.setPoints(7, u_arrow);
					break;
				}
				case PE_ArrowDown: {
					a.setPoints(7, d_arrow);
					break;
				}
				case PE_ArrowLeft: {
					a.setPoints(7, l_arrow);
					break;
				}
				case PE_ArrowRight: {
					a.setPoints(7, r_arrow);
					break;
				}
				default: {
					if (flags & Style_Up) {
						a.setPoints(7, u_arrow);
					} else {
						a.setPoints(7, d_arrow);
					}
				}
			}

			p->save();

			if (flags & Style_Down) {
				p->translate(pixelMetric(PM_ButtonShiftHorizontal),
				             pixelMetric(PM_ButtonShiftVertical));
			}

			if (flags & Style_Enabled) {
				a.translate((r.x() + r.width()/2), (r.y() + r.height()/2));
				if (p->pen() == QPen::NoPen) {
					p->setPen(cg.buttonText());
				}
				p->drawLineSegments(a, 0, 3);
				p->drawPoint(a[6]);
			} else {
				a.translate((r.x() + r.width()/2)+1, (r.y() + r.height()/2)+1);
				if (p->pen() == QPen::NoPen) {
					p->setPen(cg.highlightedText());
				}
				p->drawLineSegments(a, 0, 3);
				p->drawPoint(a[6]);

				a.translate(-1,-1);
				if (p->pen() == QPen::NoPen) {
					p->setPen(cg.mid());
				}
				p->drawLineSegments(a, 0, 3);
				p->drawPoint(a[6]);
			}

			p->restore();
			break;
		}

		default: {
			return winstyle->drawPrimitive(pe, p, r, cg, flags, opt);
		}
	}
}


void reinhardt::drawControl(ControlElement element,
                              QPainter *p,
                              const QWidget *widget,
                              const QRect &r,
                              const QColorGroup &cg,
                              SFlags flags,
                              const QStyleOption& opt) const
{
	switch (element) {
	//	CHECKBOXES
	//	----------
		case CE_CheckBox: {
			drawPrimitive(PE_Indicator, p, r, cg, flags);
			if (flags & Style_On || flags & Style_Down) {
				drawPrimitive(PE_CheckMark, p, r, cg, flags);
			}
			if (flags & Style_HasFocus) {
				p->setPen(cg.mid());
				p->setBrush(NoBrush);
				p->drawRect(r.x() + 1, r.y() + 1, r.width() - 2, r.height() - 2);
				
			}
			break;
		}

	//	RADIOBUTTONS
	//	------------
		case CE_RadioButton: {
			drawPrimitive(PE_ExclusiveIndicator, p, r, cg, flags);

			if (flags & Style_On || flags & Style_Down) {
				QCOORD center[] = { 4, 5, 4, 7, 5, 8, 7, 8, 8, 7, 8, 5, 7, 4, 5, 4 };
				QPointArray c(8, center);
				p->setPen(flags & Style_Down ? cg.highlight() : cg.text());
				p->setBrush(flags & Style_Down ? cg.highlight() : cg.text());
				p->drawPolygon(c);
			}
			if (flags & Style_HasFocus) {
				p->setPen(cg.mid());
				p->setBrush(NoBrush);
				p->drawRect(r.x() + 1, r.y() + 1, r.width() - 2, r.height() - 2);
				
			}
			break;
		}

	//	PROGRESSBAR
	//	-----------
		case CE_ProgressBarGroove: {
			//renderPanel(p, r, cg, true);
			break;
		}
		case CE_ProgressBarContents: {
			const QProgressBar* pb = (const QProgressBar*)widget;
			QRect cr = subRect(SR_ProgressBarContents, widget);
			double progress = pb->progress();
			int steps = pb->totalSteps();

			if (!cr.isValid())
				return;

			p->setPen(cg.dark());
			p->drawRect(r);

			//Make sure that cr is the right size for the contents...
			cr.setRect(cr.x() + 3, cr.y() + 3, cr.width() - 6, cr.height() - 5);

			// Draw progress bar
			if (progress > 0 || steps == 0) {
				double pg = (steps == 0) ? 0.1 : progress / steps;
				int width = QMIN(cr.width(), (int)(pg * cr.width()));
				if (steps == 0)
					width = QMIN(width,20); //Don't cross squares

				if (steps == 0) { //Busy indicator
					if (width < 1) width = 1; //A busy indicator with width 0 is kind of useless
	
					int remWidth = cr.width() - width; //Never disappear completely
					if (remWidth <= 0) remWidth = 1; //Do something non-crashy when too small...
	
					int pstep =  int(progress) % ( 2 *  remWidth );
	
					if ( pstep > remWidth ) {
						//Bounce about.. We're remWidth + some delta, we want to be remWidth - delta...
						// - ( (remWidth + some delta) - 2* remWidth )  = - (some deleta - remWidth) = remWidth - some delta..
						pstep = - (pstep - 2 * remWidth );
					}
					if( QApplication::reverseLayout() )
						p->fillRect(QRect(cr.x() + cr.width() - width - pstep, cr.y(), width, cr.height()), cg.dark());
					else
						p->fillRect(cr, cg.dark());
				return;
				}
	
	
				if( QApplication::reverseLayout() )
				{
					p->fillRect(QRect( cr.x()+(cr.width()-width), cr.y(), width, cr.height()), cg.highlight());
				}
				else
				{
					p->fillRect(QRect(cr.x(), cr.y(), width, cr.height()-1), cg.highlight());
				}
			}
			p->setPen(cg.background());
			p->drawPoint(r.topLeft());
			p->drawPoint(r.topRight());
			p->drawPoint(r.bottomLeft());
			p->drawPoint(r.bottomRight());
			break;
	  	}

		case CE_ProgressBarLabel:
		{
			const QProgressBar *progressbar = (const QProgressBar *) widget;
			QColor penColor = cg.highlightedText();
			QColor *pcolor = 0;
			if ( progressbar->centerIndicator() && !progressbar->indicatorFollowsStyle() &&
				progressbar->progress()*2 >= progressbar->totalSteps() )
				pcolor = &penColor;
			drawItem(p, r, AlignCenter | SingleLine, cg, flags & Style_Enabled, 0,
				progressbar->progressString(), -1, pcolor );
			break;
		}

	//	TABS
	//	----
		case CE_TabBarTab: {
			const QTabBar * tb = (const QTabBar *) widget;
			QTabBar::Shape tbs = tb->shape();
			bool selected = flags & Style_Selected;
			int x, y, w, h;
			r.rect(&x, &y, &w, &h);

			switch (tbs) {
				case QTabBar::RoundedAbove: {
					p->setPen(cg.background().dark());
					if (selected) {
						p->setPen(cg.button().dark());
						p->setBrush(cg.button());
						p->drawRect(r);

						p->setPen(cg.background());
						p->drawLine(x + 1, y + h - 1, x + w - 1, y + h - 1);
						p->drawPoint(r.topLeft());
						p->drawPoint(r.topRight());
					} else {
						QRect r2(x, y + 1, w, h - 1);

						p->setPen(cg.button().dark());
						useOldTabs ? p->setBrush(cg.button().dark(115)) : p->setBrush(cg.button().dark(103));
						p->drawRect(r2);

						p->setPen(cg.background());
						p->drawPoint(r2.topLeft());
						p->drawPoint(r2.topRight());
						if (useDropShadows) {
							p->setPen(cg.button().dark(153));
							p->drawLine(x + 1, y + h - 2, x + w - 2, y + h - 2);
							p->setPen(cg.button().dark(123));
							p->drawLine(x + 1, y + h - 3, x + w - 2, y + h - 3);
							p->setPen(cg.button().dark(113));
							p->drawLine(x + 1, y + h - 4, x + w - 2, y + h - 4);
						}
					}
					break;
				}

				case QTabBar::RoundedBelow: {
					p->setPen(cg.background().dark());
					if (selected) {
						p->setPen(cg.button().dark());
						p->setBrush(cg.button());
						p->drawRect(r);
						p->setPen(cg.background());
						p->drawPoint(r.bottomLeft());
						p->drawLine(x + 1, y, x + w - 2, y);
						p->drawPoint(r.bottomRight());
					} else {
						QRect r2(x, y, w, h - 1);
						p->setPen(cg.button().dark());
						useOldTabs ? p->setBrush(cg.button().dark(115)) : p->setBrush(cg.button().dark(103));
						p->drawRect(r2);

						p->setPen(cg.background());
						p->drawPoint(r2.bottomLeft());
						p->drawPoint(r2.bottomRight());
						if (useDropShadows) {
							p->setPen(cg.button().dark(153));
							p->drawLine(x + 1, y + 1, x + w - 2, y + 1);
							p->setPen(cg.button().dark(123));
							p->drawLine(x + 1, y + 2, x + w - 2, y + 2);
							p->setPen(cg.button().dark(113));
							p->drawLine(x + 1, y + 2, x + w - 2, y + 3);
						}
					}
					break;
				}

				default:
					winstyle->drawControl(element, p, widget, r, cg, flags, opt);
			}

			break;
		}

		case CE_TabBarLabel: {
			const int text_flags = AlignVCenter | AlignHCenter | ShowPrefix |
			                       DontClip | SingleLine;

			p->save();

			if (useTextShadows) {
				p->setPen(cg.background().dark(115));
				p->drawText(r.x()+1, r.y()+1, r.width(), r.height(), text_flags, opt.tab()->text());
			}

			p->setPen(flags & Style_Enabled ? cg.foreground() : cg.mid().dark());

			if (!flags & Style_Selected) {
				p->translate(pixelMetric(PM_ButtonShiftHorizontal),
				             pixelMetric(PM_ButtonShiftVertical));
			}

			p->drawText(r, text_flags, opt.tab()->text());

			if (flags & Style_HasFocus)
				drawPrimitive(PE_FocusRect, p, r, cg);

			p->restore();
			break;
		}

		case CE_PushButton: {
			QPushButton *button = (QPushButton *)widget;
			QRect br = r;
			bool btnDefault = button->isDefault();

			static int di = pixelMetric(PM_ButtonDefaultIndicator);
			br.addCoords(di, di, -di, -di);

			if (btnDefault) {
				drawPrimitive(PE_ButtonDefault, p, r, cg, flags);
			}
			
			drawPrimitive(PE_ButtonBevel, p, br, cg, flags);
			break;
		}

		case CE_PushButtonLabel: {
			const QPushButton *pb = (const QPushButton *)widget;
			const bool enabled = flags & Style_Enabled;
			const int text_flags = AlignVCenter | AlignHCenter | ShowPrefix |
			                       DontClip | SingleLine;
			QRect ur(r);
			
			if (flags & Style_Down) {
				p->translate(pixelMetric(PM_ButtonShiftHorizontal),
				             pixelMetric(PM_ButtonShiftVertical));
			}

			if (pb->text() && flags & Style_ButtonDefault) {
				p->setFont(QFont(p->font().family(), p->font().pointSize(), 75));
			}

			if (pb->iconSet() && !pb->iconSet()->isNull()) {
				QIconSet::Mode mode = enabled ? QIconSet::Normal : QIconSet::Disabled;
				QPixmap pixmap = pb->iconSet()->pixmap(QIconSet::Small, mode);
				
				if (!pb->text().isEmpty())
				{
					const int TextToIconMargin = 3;
					int length = pixmap.width() + TextToIconMargin
					              + p->fontMetrics().size(ShowPrefix, pb->text()).width();
					int offset = (r.width() - length)/2;
					
					p->drawPixmap( r.x() + offset, r.y() + r.height() / 2 - pixmap.height() / 2, pixmap );
					ur.addCoords(offset + pixmap.width() + TextToIconMargin, 0, -offset, 0);
				}
				else
				{
					if (!pb->pixmap())
						p->drawPixmap(r.x() + r.width()/2 - pixmap.width()/2, r.y() + r.height() / 2 - pixmap.height() / 2,
										pixmap);
					else  //icon + pixmap. Ugh. 
						p->drawPixmap(r.x() + pb->isDefault() ? 8 : 4 , r.y() + r.height() / 2 - pixmap.height() / 2, pixmap);
				}
			}

			if (pb->pixmap() && !pb->text()) {
				QRect pr(0, 0, pb->pixmap()->width(), pb->pixmap()->height());
				pr.moveCenter(r.center());
				p->drawPixmap(pr.topLeft(), *pb->pixmap());
			}

			if (useTextShadows && pb->text() && enabled) {
				p->setPen((flags & Style_Down ? cg.highlight().dark(135) : cg.background().dark(115)));
				p->drawText(ur.x()+1, ur.y()+1, ur.width(), ur.height(), text_flags, pb->text());
			}

			if (pb->text()) {
				p->setPen(enabled ? cg.foreground() : pb->palette().disabled().buttonText());
				p->drawText(ur, text_flags, pb->text());
			}

			if ( flags & Style_HasFocus )
				drawPrimitive( PE_FocusRect, p,
				               visualRect( subRect( SR_PushButtonFocusRect, widget ), widget ),
				               cg, flags );

			break;
		}

	//	MENUBAR ITEM (sunken panel on mouse over)
	//	-----------------------------------------
		case CE_MenuBarItem: {
			QMenuItem *mi = opt.menuItem();
			bool active  = flags & Style_Active;
			bool focused = flags & Style_HasFocus;
			const int text_flags =
				AlignVCenter | AlignHCenter | ShowPrefix | DontClip | SingleLine;

			if ( active && focused ) {
				p->setBrush(cg.highlight());
				p->setPen (cg.highlight().dark());
				p->drawRect(r);

				p->setPen(cg.mid());
				p->drawPoint(r.topLeft());
				p->drawPoint(r.bottomLeft());
				p->setPen(cg.background());
				p->drawPoint(r.topRight());
				p->drawPoint(r.bottomRight());
			}

			if (useTextShadows) {
				p->setPen((active && focused ? cg.highlight().dark(135) : cg.background().dark(115)));
				p->drawText(r.x() + 1, r.y() + 1, r.width(), r.height(), text_flags, mi->text());
			}
			p->setPen((active && focused ? cg.highlightedText() : cg.text()));
			p->drawText(r, text_flags, mi->text());
			break;
		}

	//	POPUPMENU ITEM (highlighted on mouseover)
	//	------------------------------------------
		case CE_PopupMenuItem: {
			const QPopupMenu *pum = (const QPopupMenu *) widget;
			static const int itemFrame = 0;
			static const int itemHMargin = 1;
			static const int itemVMargin = 1;
			static const int arrowHMargin = 6;
			static const int rightBorder = 12;
			const int tab = opt.tabWidth();

			int checkcol = QMAX(opt.maxIconWidth(), 26);
			p->save();
			int x, y, w, h;
			r.rect(&x, &y, &w, &h);

			QMenuItem *mi = opt.menuItem();

			if (!mi) {
				p->restore();
				return;
			}

			bool active = flags & Style_Active;
			bool disabled = !mi->isEnabled();
			bool checkable = pum->isCheckable();

			int xpos = x;
			int xm = itemFrame + checkcol + itemHMargin;

			if ( pum->erasePixmap() && !pum->erasePixmap()->isNull() ) {
				p->drawPixmap( x, y, *pum->erasePixmap(), x, y, w, h );
			} else {
				if (!QApplication::reverseLayout()) {
					p->fillRect( x, y, 22, h, cg.mid() );
					p->fillRect( x + 22, y, w - 22, h, cg.background().light() );
				} else { // i wonder why exactly +1 (diego)
					p->fillRect( w-22, y, 22+1, h, cg.mid() );
					p->fillRect( x, y, w - 22, h, cg.background().light() );
				}
			}

			if (mi->isSeparator()) {
				p->setPen(cg.mid());
				if (!QApplication::reverseLayout())
					p->drawLine(x + opt.maxIconWidth() + 6, y, x + w, y);
				else
					p->drawLine(w - (x + opt.maxIconWidth() + 6), y, x, y);
				p->restore();
				return;
			}

			if (active && !disabled) {
				p->setPen(cg.highlight().dark());
				p->setBrush(cg.highlight());
				p->drawRect(x + 1, y + 1, w - 2, h - 2);
				p->setPen(cg.mid());
				p->drawPoint(QPoint(x + 1, y + 1));
				p->drawPoint(QPoint(x + 1, y + h - 2));
				p->setPen(cg.background());
				p->drawPoint(QPoint(x + w - 2, y + 1));
				p->drawPoint(QPoint(x + w - 2, y + h - 2));
			}

			if (mi->iconSet()) {
				p->save();
				QIconSet::Mode mode =
					disabled ? QIconSet::Disabled : QIconSet::Normal;
				QPixmap pixmap = mi->iconSet()->pixmap(QIconSet::Small, mode);
				int pixw = pixmap.width();
				int pixh = pixmap.height();

				QRect cr(xpos, y, opt.maxIconWidth(), h);
				QRect pmr(0, 0, pixw, pixh);
				pmr.moveCenter(cr.center());

				if ( QApplication::reverseLayout() )
					pmr = visualRect( pmr, r );

				p->setPen(cg.highlightedText());
				p->drawPixmap(pmr.topLeft(), pixmap);
				p->restore();
			} else if (checkable) {
				if (mi->isChecked()) {
					int xp = xpos;

					SFlags cflags = Style_Default;
					if (!disabled)
						cflags |= Style_Enabled;
					else
						cflags |= Style_On;

					if (active)
						p->setPen(cg.highlightedText());
					else
						p->setPen(cg.buttonText());

		
					QRect rr = QRect( xp, y, 22, h );
					if ( QApplication::reverseLayout() )
						rr = visualRect( rr, r );

					drawPrimitive(PE_CheckMark, p, rr, cg, cflags);
				}
			}

			p->setPen(active ? cg.highlightedText() : cg.buttonText());

			xpos += xm;

			if (mi->custom()) {
				int m = itemVMargin;
				p->save();
				p->setPen(cg.text());
				mi->custom()->paint(p, cg, active, !disabled, x+xm, y+m, w-xm-tab+1, h-2*m);
				p->restore();
				return;
			} else {
				QString s = mi->text();
				if(!s.isNull()) {
					int t = s.find('\t');
					int m = itemVMargin;

					int text_flags =
						AlignVCenter | ShowPrefix | DontClip | SingleLine;

					if (active && !disabled) {
						p->setPen(cg.highlightedText());
					} else if (disabled) {
						p->setPen(cg.mid().light(110));
					} else {
						p->setPen(cg.text());
					}

					if (t >= 0) {
						int xp;
						xp = x + w - tab - rightBorder - itemHMargin - itemFrame + 1;

						QRect rr = QRect( xp, y+m, tab, h-2*m );
						if ( QApplication::reverseLayout() )
							rr = visualRect( rr, r );

						if (useTextShadows) {
							p->save();
							p->setPen(active && !disabled ? cg.highlight().dark(130) : cg.background().dark(115));
							p->drawText(rr.x() + 1, rr.y() + 1, rr.width(), rr.height(), text_flags, s.mid(t+1));
							p->restore();
						}

						p->drawText(rr, text_flags, s.mid(t+1));
						s = s.left(t);
					}
					
					QRect rr = QRect( xpos, y+m, w-xm-tab+1, h-2*m );
					if ( QApplication::reverseLayout() ){
						rr = visualRect( rr, r );
						text_flags |= AlignRight;
					}

					if (useTextShadows) {
						p->save();
						p->setPen(active && !disabled ? cg.highlight().dark(130) : cg.background().dark(115));
						p->drawText(rr.x() + 1, rr.y() + 1, rr.width(), rr.height(), text_flags, s);
						p->restore();
					}

					p->drawText(rr, text_flags, s);
				} else if (mi->pixmap()) {
					QPixmap *pixmap = mi->pixmap();
					if (pixmap->depth() == 1) {
						p->setBackgroundMode(OpaqueMode);
					}
					p->drawPixmap(xpos, y, *pixmap);
					if (pixmap->depth() == 1) {
						p->setBackgroundMode(TransparentMode);
					}
				}
			}

			if (mi->popup()) {
				int dim = pixelMetric(PM_MenuButtonIndicator);

				xpos = x+w - arrowHMargin - 2*itemFrame - dim;

				if (active && !disabled)
					p->setPen(cg.highlightedText());
				else
					p->setPen(cg.text());


				QRect rr = QRect(xpos, y + h/2 - dim/2, dim, dim);
				if ( QApplication::reverseLayout() )
					rr = visualRect( rr, r );
				drawPrimitive((QApplication::reverseLayout() ? PE_ArrowLeft : PE_ArrowRight), p, rr, cg, Style_Enabled);
			}

			p->restore();
			break;
		}

		default:
			winstyle->drawControl(element, p, widget, r, cg, flags, opt);
	}
}

void reinhardt::drawControlMask(ControlElement element,
                                  QPainter *p,
                                  const QWidget *w,
                                  const QRect &r,
                                  const QStyleOption &opt) const
{
	switch (element) {
		case CE_PushButton: {
			int x1, y1, x2, y2;
			r.coords (&x1, &y1, &x2, &y2);
			QCOORD corners[] = { x1, y1, x2, y1, x1, y2, x2, y2 };
			p->fillRect (r, color1);
			p->setPen (color0);
			p->drawPoints (QPointArray (4, corners));
			break;
		}
		default: {
			KStyle::drawControlMask (element, p, w, r, opt);
		}
	}
}

void reinhardt::drawComplexControl(ComplexControl control,
                                     QPainter *p,
                                     const QWidget *widget,
                                     const QRect &r,
                                     const QColorGroup &cg,
                                     SFlags flags,
                                     SCFlags controls,
                                     SCFlags active,
                                     const QStyleOption& opt) const
{
	switch(control) {
	//	LISTVIEW
	//	--------
		case CC_ListView:
			{
			if ( useOldTreeview ) {
				KStyle::drawComplexControl(control, p, widget, r, cg, flags, controls, active, opt);
				break;
			}
			if ( controls & SC_ListView ) {
				QCommonStyle::drawComplexControl( control, p, widget, r, cg, flags, controls, active, opt );
				/*QListView *listview = (QListView*)widget;
				p->fillRect( r, listview->viewport()->backgroundBrush() );
				return;*/
			}
			if ( controls & ( SC_ListViewBranch | SC_ListViewExpand ) ) {
				if (opt.isDefault())
				break;

				QListViewItem *item = opt.listViewItem(),
					*child = item->firstChild();

				int y = r.y();
				int c;
				if ( active == SC_All && controls == SC_ListViewExpand ) {
					c = 2;
				} else {
					int linebot = 0;

					// skip the stuff above the exposed rectangle
					while ( child && y + child->height() <= 0 ) {
						y += child->totalHeight();
						child = child->nextSibling();
					}

					int bx = r.width() / 2;

					// paint stuff in the magical area
					QListView* v = item->listView();
					while ( child && y < r.height() ) {
						int lh;
						if ( !item->multiLinesEnabled() )
						lh = child ? child->height() : 0;
						else
						lh = p->fontMetrics().height() + 2 * v->itemMargin();
						lh = QMAX( lh, QApplication::globalStrut().height() );
						if ( lh % 2 > 0 )
						lh++;
						linebot = y + lh/2;
						if ( (child->isExpandable() || child->childCount()) && (child->height() > 0) ) {
							p->setPen(cg.dark());
							QPointArray a;
							if ( !child->isOpen() ) { // closed
								a.setPoints(7, r_arrow);
							} else { //or open
								a.setPoints(7, d_arrow);
							}
							a.translate(bx, linebot);
							p->drawLineSegments(a, 0, 3);
							p->drawPoint(a[6]);
						}
						y += child->totalHeight();
						child = child->nextSibling();
					}
				}
				p->setPen(cg.text());
			}
			break;
		}

	//	COMBOBOX
	//	--------
		case CC_ComboBox: {
			int x, y, w, h;
			const QComboBox *cb = (const QComboBox *)widget;
			r.rect(&x, &y, &w, &h);

			if (active & Style_Sunken)
				flags |= Style_Sunken;

			static const unsigned int handle_width = 15;
			static const unsigned int handle_offset = handle_width + 1;
			int x1 = w - (handle_width / 2) - 3, y1 = r.center().y() - 4;
			const QCOORD anarrow[] = { x1, y1, x1 + 3, y1 + 3, x1 - 3, y1 + 3, x1, y1};
			y1 = r.center().y() + 4;
			const QCOORD anarrow2[] = { x1, y1, x1 - 3, y1 - 3, x1 + 3, y1 - 3, x1, y1};
			const QPointArray a(4, anarrow);
			const QPointArray a2(4, anarrow2);

			p->save();

			renderButton(p, r, cg);

			if (useOldCombo) {
				p->fillRect(x + 1, y + 1, w - 2, h - 2, cg.base());
			} else {
				cb->editable() ? p->fillRect(x + 1, y + 1, w - 2, h - 2, cg.base()) : p->fillRect(x + 1, y + 1, w - 2, h - 2, cg.button());
			}

			p->setBrush(NoBrush);
			QRect rr = QRect(x + 1, y + 1, w - 2, h - 2);

			if (QApplication::reverseLayout()) { rr = visualRect( rr, r ); }

			if (flags & Style_HasFocus || (cb->listBox() && cb->listBox()->isVisible())) {
				cb->editable() ? p->setPen(cg.base().dark(150)) : p->setPen(cg.button().dark(150));
				p->drawRect(rr);
				if (useDropShadows) {
					rr.setRect(x + 2, y + 2, w - 4, h - 4);
					cb->editable() ? p->setPen(cg.base().dark(110)) : p->setPen(cg.button().dark(110));
					p->drawRect(rr);
					QRect hr = QRect(w - handle_offset - 1, y + 1, handle_width, h - 2);
					if (QApplication::reverseLayout()) { hr = visualRect(hr, r); }
					p->drawLine(hr.x() - 1, hr.y() + 1, hr.x() - 1, hr.height() - 2);
					cb->editable() ? p->setPen(cg.base().dark(105)) : p->setPen(cg.button().dark(105));
					p->drawLine(hr.x(), hr.y() + 2, hr.x(), hr.height() - 2);
				}

			} else {
				cb->editable() ? p->setPen(cg.base().dark(110)) : p->setPen(cg.button().dark(110));
				p->drawRect(rr);
			}

		/*	Draw the arrow */
			p->setPen(flags & Style_Down ? cg.highlight() : cg.text());
			p->setBrush(flags & Style_Down ? cg.highlight() : cg.text());
			p->drawPolygon(a);
			p->drawPolygon(a2);

			p->restore();
			break;
		}

	//	ScrollBar
	//	---------
		case CC_ScrollBar:
		{
			const QScrollBar *scrollbar = (const QScrollBar *) widget;
			QRect addline, subline, addpage, subpage, slider, first, last;
			bool maxedOut = (scrollbar->minValue() == scrollbar->maxValue());

			subline = querySubControlMetrics(control, widget, SC_ScrollBarSubLine, opt);
			addline = querySubControlMetrics(control, widget, SC_ScrollBarAddLine, opt);
			subpage = querySubControlMetrics(control, widget, SC_ScrollBarSubPage, opt);
			addpage = querySubControlMetrics(control, widget, SC_ScrollBarAddPage, opt);
			slider  = querySubControlMetrics(control, widget, SC_ScrollBarSlider,  opt);
			first   = querySubControlMetrics(control, widget, SC_ScrollBarFirst,   opt);
			last    = querySubControlMetrics(control, widget, SC_ScrollBarLast,    opt);

			if ((controls & SC_ScrollBarSubLine) && subline.isValid())
				drawPrimitive(PE_ScrollBarSubLine, p, subline, cg, Style_Enabled | ((active == SC_ScrollBarSubLine) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
			if ((controls & SC_ScrollBarAddLine) && addline.isValid())
				drawPrimitive(PE_ScrollBarAddLine, p, addline, cg, Style_Enabled | ((active == SC_ScrollBarAddLine) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
			if ((controls & SC_ScrollBarSubPage) && subpage.isValid())
				drawPrimitive(PE_ScrollBarSubPage, p, subpage, cg, Style_Enabled | ((active == SC_ScrollBarSubPage) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
        		if ((controls & SC_ScrollBarAddPage) && addpage.isValid())
				drawPrimitive(PE_ScrollBarAddPage, p, addpage, cg, ((maxedOut) ? Style_Default : Style_Enabled) | ((active == SC_ScrollBarAddPage) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
        		if ((controls & SC_ScrollBarFirst) && first.isValid())
				drawPrimitive(PE_ScrollBarFirst, p, first, cg, Style_Enabled | ((active == SC_ScrollBarFirst) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
			if ((controls & SC_ScrollBarLast) && last.isValid())
				drawPrimitive(PE_ScrollBarLast, p, last, cg, Style_Enabled | ((active == SC_ScrollBarLast) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));
			if ((controls & SC_ScrollBarSlider) && slider.isValid()) {
				drawPrimitive(PE_ScrollBarSlider, p, slider, cg, Style_Enabled | ((active == SC_ScrollBarSlider) ? Style_Down : Style_Default) | ((scrollbar->orientation() == Qt::Horizontal) ? Style_Horizontal : 0));

				// ### perhaps this should not be able to accept focus if maxedOut?
				if (scrollbar->hasFocus()) {
					QRect fr(slider.x() + 2, slider.y() + 2, slider.width() - 5, slider.height() - 5);
					drawPrimitive(PE_FocusRect, p, fr, cg, Style_Default);
				}

        		}


			QRect cr(1, 1, 2, 2);
			p->setPen(cg.background().dark());
			if ( scrollbar->orientation() == Qt::Horizontal ) {
				cr.setRect(r.x() + subline.width(), r.y(), r.width() - (subline.width() + addline.width() ), r.height());
				p->drawRect(cr);
			} else {
				cr.setRect(r.x(), r.y() + subline.height(), r.width(), r.height() - (subline.height() + addline.height() ));
				p->drawRect(cr);
			}
		break;
		}

	//	TOOLBUTTON
	//	----------
		case CC_ToolButton: {
			const QToolButton *tb = (const QToolButton *) widget;

			QRect button, menuarea;
			button   = querySubControlMetrics(control, widget, SC_ToolButton, opt);
			menuarea = querySubControlMetrics(control, widget, SC_ToolButtonMenu, opt);

			SFlags bflags = flags,
			       mflags = flags;

			if (active & SC_ToolButton)
				bflags |= Style_Down;

			if (active & SC_ToolButtonMenu)
				mflags |= Style_Down;

			if (controls & SC_ToolButton) {
			//	If we're pressed, on, or raised...
				if (bflags & (Style_Down | Style_On | Style_Raised)) {
					drawPrimitive(PE_ButtonTool, p, button, cg, bflags, opt);
				} else if (tb->parentWidget() &&
				           tb->parentWidget()->backgroundPixmap() &&
				           !tb->parentWidget()->backgroundPixmap()->isNull()) {
					QPixmap pixmap = *(tb->parentWidget()->backgroundPixmap());
					p->drawTiledPixmap( r, pixmap, tb->pos() );
				}
			}

			// Draw a toolbutton menu indicator if required
			if (controls & SC_ToolButtonMenu) {
				if (mflags & (Style_Down | Style_On | Style_Raised)) {
					drawPrimitive(PE_ButtonDropDown, p, menuarea, cg, mflags, opt);
				}
				drawPrimitive(PE_ArrowDown, p, menuarea, cg, mflags, opt);
			}

			if (tb->hasFocus() && !tb->focusProxy()) {
				QRect fr = tb->rect();
				fr.addCoords(2, 2, -2, -2);
				drawPrimitive(PE_FocusRect, p, fr, cg);
			}

			break;
		}

	//	SPINWIDGETS
	//	-----------
		case CC_SpinWidget: {
			const QSpinWidget *sw = (const QSpinWidget *) widget;
			SFlags flags;
			PrimitiveElement pe;

			if (controls & SC_SpinWidgetFrame) {
				renderButton(p, r, cg);
				p->fillRect(r.x() + 1, r.y() + 1, r.width() - 2, r.height() - 2, cg.base());
			}

			if (controls & SC_SpinWidgetUp) {
				flags = Style_Default | Style_Enabled;
				if (active == SC_SpinWidgetUp) {
					flags |= Style_On;
					flags |= Style_Sunken;
				} else
					flags |= Style_Raised;

				if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
					pe = PE_SpinWidgetPlus;
				else
					pe = PE_SpinWidgetUp;

				QRect re = sw->upRect();
				QColorGroup ucg = sw->isUpEnabled() ? cg : sw->palette().disabled();
				p->fillRect(re, QBrush(cg.base()));
				drawPrimitive(pe, p, re, ucg, flags);
			}

			if (controls & SC_SpinWidgetDown) {
				flags = Style_Default | Style_Enabled;
				if (active == SC_SpinWidgetDown) {
					flags |= Style_On;
					flags |= Style_Sunken;
				} else
					flags |= Style_Raised;

				if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
					pe = PE_SpinWidgetMinus;
				else
					pe = PE_SpinWidgetDown;

				QRect re = sw->downRect();
				QColorGroup dcg = sw->isDownEnabled() ? cg : sw->palette().disabled();
				p->fillRect(re, QBrush(cg.base()));
				drawPrimitive(pe, p, re, dcg, flags);
			}
				p->setPen(cg.dark());
				if (sw->hasFocus()) p->drawRect(r.x() + 1, r.y() + 1, r.width() - 2, r.height() - 2);
			break;
		}

  default:
			KStyle::drawComplexControl(control, p, widget,
			                           r, cg, flags, controls,
			                           active, opt);
			break;
	}
}


QRect reinhardt::subRect(SubRect r, const QWidget *widget) const
{
// Don't use KStyles progressbar subrect
	switch (r) {
		case SR_ComboBoxFocusRect: {
			return querySubControlMetrics( CC_ComboBox, widget, SC_ComboBoxEditField );
		}

		case SR_PushButtonFocusRect: {
			QRect rect = widget->rect();
			int margin = pixelMetric(PM_ButtonDefaultIndicator, widget) + pixelMetric(PM_DefaultFrameWidth, widget);

			rect.addCoords(margin, margin, -margin, -margin);
			return rect;
		}
		
		case SR_ProgressBarGroove:
		case SR_ProgressBarContents:
		case SR_ProgressBarLabel: {
			return winstyle->subRect(r, widget);
		}
		default: {
			return KStyle::subRect(r, widget);
		}
	}
}

QRect reinhardt::querySubControlMetrics(ComplexControl control,
                                          const QWidget *widget,
                                          SubControl subcontrol,
                                          const QStyleOption &opt) const
{
	if (!widget) {
		return QRect();
	}

	QRect r(widget->rect());
	switch (control) {
		case CC_ComboBox: {
			switch (subcontrol) {
				case SC_ComboBoxEditField: {
					return QRect(r.x() + 2, r.y() + 2, r.width() - 19, r.height() - 4);
				}
				default: {
					return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
				}
			}
			break;
		}
		default: {
		}
	}
	return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
}

int reinhardt::pixelMetric(PixelMetric m, const QWidget *widget) const
{
	switch(m) {
	//	TABS
	//	----
		case PM_TabBarTabVSpace: {
			const QTabBar * tb = (const QTabBar *) widget;
			if (tb->shape() == QTabBar::RoundedAbove ||
			    tb->shape() == QTabBar::RoundedBelow)
				return 10;
			else
				return 0;
		}

		case PM_TabBarTabOverlap: {
			const QTabBar* tb = (const QTabBar*)widget;
			if (tb->shape() != QTabBar::RoundedAbove &&
			    tb->shape() != QTabBar::RoundedBelow)
				return 3;                   // Leave standard size alone
			else
				return 1;                   // Change size for our tabs only
		}

	//	SPLITTERS
	//	---------
		case PM_SplitterWidth: {
			return 7;
		}

	//	PROGRESSBARS
	//	------------
		case PM_ProgressBarChunkWidth:
			return 9;

	//	SLIDER
	//	------
		case PM_SliderLength:
			return 11;

	//	MENU INDICATOR
	//	--------------
		case PM_MenuButtonIndicator:
			return 8;

	//	HEADER INDICATOR
	//	----------------
		case PM_HeaderMarkSize:
			return 1;

	//	CHECKBOXES / RADIO BUTTONS
	//	--------------------------
		case PM_ExclusiveIndicatorWidth:    // Radiobutton size
		case PM_ExclusiveIndicatorHeight:   // 13x13
		case PM_IndicatorWidth:             // Checkbox size
		case PM_IndicatorHeight:            // 13x13
			return 13;

	//	FRAMES
	//	------
		case PM_SpinBoxFrameWidth:
			return 1;

		case PM_MenuBarFrameWidth:
			return 0;

		case PM_DefaultFrameWidth: {
			if (widget &&
			   (widget->inherits("QPopupMenu") ||
			   widget->inherits("QMenuBar") ||
			   widget->inherits("QLineEdit") ||
			   widget->inherits("QRangeControl") ||
			   widget->inherits("QScrollView"))) {
				return 1;
			} else {
				return 2;
			}
		}
		
		case PM_ButtonDefaultIndicator: {
			return 1;
		}

		case PM_ButtonShiftVertical:
		case PM_ButtonShiftHorizontal: {
			return 1;
		}

		default:
			return KStyle::pixelMetric(m, widget);
	}
}


QSize reinhardt::sizeFromContents(ContentsType t,
                                    const QWidget *widget,
                                    const QSize &s,
                                    const QStyleOption &opt) const
{
	switch (t) {
		case CT_MenuBar: {
			const QMenuBar *mb = (const QMenuBar *)widget;
			int w, h;
			{
				QFont font = mb->font();
				h = QFontInfo(font).pixelSize() + 6;
				w = 10;
			}

			return QSize(w, h);

			break;
		}
		case CT_PopupMenuItem: {
		//	const QPopupMenu *popup = (const QPopupMenu *)widget;
			QMenuItem *mi = opt.menuItem();
			if (!widget || opt.isDefault())
				break;

			int w = s.width(), h = s.height();

			if (mi->custom()) {
				w = mi->custom()->sizeHint().width();
				h = mi->custom()->sizeHint().height();
			}
			else if (mi->widget()) {
				// don't change the size in this case.
			}
			else if (mi->isSeparator()) {
				w = 10;
				h = 1;
			} else {
				return winstyle->sizeFromContents(t, widget, s, opt);
			}

			return QSize(w, h);
		}

		case CT_PushButton: {
			const QPushButton* button = (const QPushButton*) widget;
			int w  = s.width();
			int h  = s.height();
			int bm = pixelMetric( PM_ButtonMargin, widget );
			int fw = pixelMetric( PM_DefaultFrameWidth, widget ) * 2;

			//We use bold font sometimes, so that make take more space than Qt allocated.
			QFontMetrics origFM = button->fontMetrics();
			QFontMetrics boldFM = QFontMetrics(QFont(button->font().family(), button->font().pointSize(), 75));
			int origTextWidth = origFM.size(ShowPrefix, button->text()).width();
			int boldTextWidth = boldFM.size(ShowPrefix, button->text()).width();

			if (boldTextWidth > origTextWidth) //this might not happen in particular when text is empty
				w += boldTextWidth - origTextWidth;
			
			w += bm + fw;
			h += bm + fw;

			// Ensure we stick to standard width and heights.
			if ( button->isDefault() || button->autoDefault() ) {
				if ( w < 80 && !button->pixmap() )
					w = 80;
			}
				
			// Compensate for default indicator. Doing it for all widgets lets us
			// workaround changes in auto-default behavior in KDialogBase in 3.3
			int di = pixelMetric( PM_ButtonDefaultIndicator );
			w += di * 2;
			h += di * 2;

			if ( h < 20 )
				h = 20;

			return QSize( w, h );
		}
		
		case CT_ToolButton:
		{
			if(widget->parent() && ::qt_cast<QToolBar*>(widget->parent()) )
				return QSize( s.width()+2*4, s.height()+2*4 );
			else
				return KStyle::sizeFromContents (t, widget, s, opt);
		}

		case CT_ComboBox: {
			int arrow = 21;
			return QSize( s.width() + arrow, QMAX(s.height() + 4, 16));
		}

		default:
			return KStyle::sizeFromContents (t, widget, s, opt);
	}

	return KStyle::sizeFromContents (t, widget, s, opt);
//	return winstyle->sizeFromContents(t, widget, s, opt);
}

void reinhardt::paletteChanged()
{
	QComboBox *combo   = 0L;
	QButton   *button  = 0L;
	QMenuBar  *menuBar = 0L;
	QToolBar  *toolBar = 0L;
	QWidget   *widget  = 0L;

	QValueListIterator<QWidget*> it = m_widgets.begin();
	for ( ; it != m_widgets.end(); ++it ) {
		widget = *it;
		if ((combo = dynamic_cast<QComboBox*>( widget )))
			updatePalette( combo );
		else if ((button = dynamic_cast<QButton*>( widget)))
			updatePalette( button );
		else if ( (toolBar = dynamic_cast<QToolBar*>( widget )))
			updatePalette( toolBar );
		else if ((menuBar = dynamic_cast<QMenuBar*>( widget )))
			updatePalette( menuBar );
	}
}

void reinhardt::updatePalette( QComboBox *combo )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Button));
	//pal.setColor(QColorGroup::Dark, pal.active().color(QColorGroup::Base));
	//pal.setColor(QColorGroup::Midlight, pal.active().color(QColorGroup::Background));
	combo->setPalette(pal);
}

void reinhardt::updatePalette( QButton *button )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Button));
	button->setPalette(pal);
}

void reinhardt::updatePalette( QCheckBox *checkbox )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Background, pal.active().color(QColorGroup::Background));
	checkbox->setPalette(pal);
}

void reinhardt::updatePalette( QRadioButton *radiobutton )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Background, pal.active().color(QColorGroup::Background));
	radiobutton->setPalette(pal);
}

void reinhardt::updatePalette( QToolBar *bar )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Background));
	bar->setPalette(pal);
}

void reinhardt::updatePalette( QMenuBar *bar )
{
	QPalette pal = QApplication::palette();
	pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Background));
	bar->setPalette(pal);
}

void reinhardt::slotDestroyed()
{
	m_widgets.remove( (QWidget*) sender() );
}

bool reinhardt::eventFilter(QObject *obj, QEvent *ev)
{
    if (useHighlightedButtons) {
		if (obj->inherits("QButton") || obj->inherits("QComboBox")) {
		if ( ! ::qt_cast<QToolButton*>(obj) ) {
    			if (ev->type() == QEvent::Enter) {
				QWidget *btn = (QWidget *)obj;
				if (btn->isEnabled()) {
					QPalette pal = QApplication::palette();
					pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Button).light(115));
					btn->setPalette(pal);
				}
			} else if (ev->type() == QEvent::Leave) {
				QWidget *btn = (QWidget *)obj;
				QPalette pal = QApplication::palette();
				pal.setColor(QColorGroup::Button, pal.active().color(QColorGroup::Button));
				btn->setPalette(pal);
			}
		}
		} else if (obj->inherits("QCheckBox") || obj->inherits("QRadioButton")) {
			if (ev->type() == QEvent::Enter) {
				QWidget *btn = (QWidget *)obj;
				if (btn->isEnabled()) {
					QPalette pal = QApplication::palette();
					pal.setColor(QColorGroup::Background, pal.active().color(QColorGroup::Background).light(115));
					btn->setPalette(pal);
				}
			} else if (ev->type() == QEvent::Leave) {
				QWidget *btn = (QWidget *)obj;
				QPalette pal = QApplication::palette();
				pal.setColor(QColorGroup::Background, pal.active().color(QColorGroup::Background));
				btn->setPalette(pal);
			}
		}
	}
	return false;
}

#include "reinhardt.moc"

// vim: noet ts=4 sw=4
